#include <iostream>
#include <cstdlib>
#include <cmath>
#include "config.h"
using namespace std;

const int nHistoBins = 50;

static double fun(double x)
	{
	return exp(x)/( pow(sin(x),3.0) + pow(cos(x),3.0) );
	}

static double d1(double h, double x0)
	{
	double x[] = {x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h};
	double y[] = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]) };
	return (y[0]/12.0-2.0*y[1]/3.0+2.0*y[3]/3.0-y[4]/12.0)/pow(h,1);
	}

static double d2(double h, double x0)
	{
	double x[] = {x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h};
	double y[] = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]) };
	return (-y[0]/12.0+4.0*y[1]/3.0-5.0*y[2]/2.0+4.0*y[3]/3.0-y[4]/12.0)/pow(h,2);
	}

static double d3(double h, double x0)
	{
	double x[] = {x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h};
	double y[] = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6]) };
	return (y[0]/8.0-y[1]+13.0*y[2]/8.0-13.0*y[4]/8.0+y[5]-y[6]/8.0)/pow(h,3);
	}

static double d4(double h, double x0)
	{
	double x[] = {x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h};
	double y[] = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6]) };
	return (-y[0]/6.0+2.0*y[1]-13.0*y[2]/2.0+28.0*y[3]/3.0-13.0*y[4]/2.0+2.0*y[5]-y[6]/6.0)/pow(h,4);
	}

static int binNb(double xMin, double xMax, int nBins, double x)
		{
		double dx = (xMax-xMin)/nBins;
		double delta = x-xMin;
		return (int)floor(delta/dx);
		}

static void createHisto(double histo[2][nHistoBins], double xMin,double xMax)
	{
	double dx = (xMax-xMin)/nHistoBins;
	histo[0][0] = xMin + dx/2.0;

	for (int i=0; i<nHistoBins; i++)
		{
		histo[0][i] = histo[0][0] + i*dx;
		histo[1][i] = 0.0;
		}
	}

static void fillHisto(double histo[2][nHistoBins], double x)
	{
	int L  = sizeof(histo[0]) / sizeof(histo[0][0]);
	double dx = (histo[0][L-1]-histo[0][0])/(L-1.0);
	double xMin = histo[0][0]-dx/2.0;
	double xMax = histo[0][L-1]+dx/2.0;
	int bin = binNb(xMin,xMax,L,x);
	if (0<=bin && bin<L) histo[1][bin]++;
	}

static void normHisto(double histo[2][nHistoBins])
	{
	int L  = sizeof(histo[0]) / sizeof(histo[0][0]);
	double sum = 0.0;
	for (int i=0; i<L; i++) sum = sum + histo[1][i];
	for (int i=0; i<L; i++) histo[1][i] = histo[1][i]/sum;
	}

int main(int argc, char **argv)
	{
	static double stat = 1E7;
	static bool h_isDisturbed=false;

	static double dx = 1E-11;

	static double h0_d1 = 1E-4;
	static double h0_d2 = 1E-3;
	static double h0_d3 = 1E-3;
	static double h0_d4 = 1E-2;

	static double x0 = 0;

	static double truth_d1 = 1;
	static double truth_d2 = 4;
	static double truth_d3 = 4;
	static double truth_d4 = 28;

	double ranNb;

	double truth = 4.0;
	double width = 0.2E-9;
	double xMin = truth-width;
	double xMax = truth+width;
	double histo[2][nHistoBins];
	createHisto(histo, xMin, xMax);

	double kahanSum[4];
	double kahanC[4];
	for (int i=0; i<4; i++)
			{
			kahanSum[i] = 0.0;
			kahanC[i] = 0.0;
			}
	double di[4];
	double kahanY[4];
	double kahanT[4];

	srand(time(NULL));
	for (int i=0; i<stat; i++)
		{
		double x = x0;
		double h;

		if (h_isDisturbed)
			{
			ranNb=((double)rand()/(double)RAND_MAX);
			h = h0_d1 *( ranNb+0.5 );
			di[0] = d1(h,x);

			ranNb=((double)rand()/(double)RAND_MAX);
			h = h0_d2 *( ranNb+0.5 );
			di[1] =  d2(h,x);

			ranNb=((double)rand()/(double)RAND_MAX);
			h = h0_d3 *( ranNb+0.5 );
			di[2] = d3(h,x);

			ranNb=((double)rand()/(double)RAND_MAX);
			h = h0_d4 *( ranNb+0.5 );
			di[3] = d4(h,x);

			for (int j=0; j<4; j++)
				{
				kahanY[j] = di[j] - kahanC[j];
				kahanT[j] = kahanSum[j] + kahanY[j];
				kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j];
				kahanSum[j] = kahanT[j];
				}

			fillHisto(histo,di[1]);
			}
		else
			{
			ranNb=((double)rand()/(double)RAND_MAX);
			x = x0 + dx*( 2.0*ranNb -1.0);

			di[0] = d1(1E-4,x);
			di[1] = d2(1E-3,x);
			di[2] = d3(1E-3,x);
			di[3] = d4(1E-2,x);

			for (int j=0; j<4; j++)
				{
				kahanY[j] = di[j] - kahanC[j];
				kahanT[j] = kahanSum[j] + kahanY[j];
				kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j];
				kahanSum[j] = kahanT[j];
				}

			fillHisto(histo,di[1]);
			}

		}

	double df1_si = d1(1E-4,x0);
	double d1_si = abs(truth_d1 -df1_si);
	double df1_av = kahanSum[0]/stat;
	double d1_av = abs(truth_d1 -df1_av);

	double df2_si = d2(1E-3,x0);
	double d2_si = abs(truth_d2 -df2_si);
	double df2_av = kahanSum[1]/stat;
	double d2_av = abs(truth_d2 -df2_av);

	double df3_si = d3(1E-3,x0);
	double d3_si = abs(truth_d3 -df3_si);
	double df3_av = kahanSum[2]/stat;
	double d3_av = abs(truth_d3 -df3_av);

	double df4_si = d4(1E-2,x0);
	double d4_si = abs(truth_d4 -df4_si);
	double df4_av = kahanSum[3]/stat;
	double d4_av = abs(truth_d4 -df4_av);

	std::cout << "" << std::endl;
	std::cout << "First Derivative" << std::endl;
	std::cout << "Truth = " << truth_d1 << std::endl;
	std::cout << "AV = " << df1_av << ", delta = " <<  d1_av  << std::endl;
	std::cout << "Single = " << df1_si << ", delta = " << d1_si  << std::endl;
	std::cout << "Sing/AV delta  = " << d1_si/d1_av << std::endl;

	std::cout << "" << std::endl;
	std::cout << "Second Derivative" << std::endl;
	std::cout << "Truth = " << truth_d2 << std::endl;
	std::cout << "AV = " << df2_av << ", delta = " <<  d2_av  << std::endl;
	std::cout << "Single = " << df2_si << ", delta = " << d2_si  << std::endl;
	std::cout << "Sing/AV delta  = " << d2_si/d2_av << std::endl;

	std::cout << "" << std::endl;
	std::cout << "Third Derivative" << std::endl;
	std::cout << "Truth = " << truth_d3 << std::endl;
	std::cout << "AV = " << df3_av << ", delta = " <<  d3_av  << std::endl;
	std::cout << "Single = " << df3_si << ", delta = " << d3_si  << std::endl;
	std::cout << "Sing/AV delta  = " << d3_si/d3_av << std::endl;

	std::cout << "" << std::endl;
	std::cout << "Fourth Derivative" << std::endl;
	std::cout << "Truth = " << truth_d4 << std::endl;
	std::cout << "AV = " << df4_av << ", delta = " <<  d4_av  << std::endl;
	std::cout << "Single = " << df4_si << ", delta = " << d4_si  << std::endl;
	std::cout << "Sing/AV delta  = " << d4_si/d4_av << std::endl;

	std::cout << "" << std::endl;
	std::cout << "HISTOS" << std::endl;
	std::cout << "" << std::endl;

	normHisto(histo);
		std::cout.precision(5);
		std::cout << "" << std::endl;
		for (int i=0; i<nHistoBins; i++)
			{
			double xVal = histo[0][i]-4.0;
			double yVal = histo[1][i];
			std::cout <<  "(" << xVal << "," << yVal << ") "  << " ";
			}

	std::cout << "" << std::endl;

	return 0;
	}
