#include <iostream>
#include "config.h"
#include<math.h>
#include <time.h>

double determinant(double [][25], double);
void cofactor(double [][25], double, double [][25]);
void transpose(double [][25], double [][25], double, double [][25]);

static double stat = 1E5;
static double eps = 1E-12; // Iny krok vhodny pre ine koeficienty -12 -13 -14
static double cfs[] = {1.0, -1.0 / 3.0, 1.0 / 7.0, -1.0 / 11.0};

static double fn(double x)
	{
	return cfs[0] + cfs[1] * cos(x) + cfs[2] * cos(2.0 * x) + cfs[3] * cos(3.0 * x);
	}

static void prod(double inMat[25][25], double inVec[25], double outVec[25])
	{
	int i,j;
	for (i=0; i<25; i++) outVec[i]=0;

	for (i=0; i<25; i++)
		{
		for (j=0; j<25; j++) outVec[i] = outVec[i] + inMat[i][j]*inVec[j];
		}
	}

int main(int argc, char **argv)
	{
	double d = 1E-3;
	double x[] = {1.0,1.0+d,1.0+d+d,1.0+d+d+d};
	double y[25];
	for (int i=0; i<25; i++) y[i] = 0.0;
	for (int i=0; i<4; i++) y[i] = fn(x[i]);

	double M[25][25], iM[25][25];
	for (int i=0; i<25; i++)
		{
		for (int j=0; j<25; j++)
			{
			M[i][j]=0.0;
			iM[i][j]=0.0;
			}
		}

	int dim = 4;
	for (int i=0; i<4; i++)
		{
		M[i][0] = 1.0;
		M[i][1] = cos( x[i] );
		M[i][2] = cos( 2.0*x[i] );
		M[i][3] = cos( 3.0*x[i] );
		}

	cofactor(M,dim,iM);

	double cfs_single[25];
	prod(iM,y,cfs_single);

	double kahanSum[4];
	double kahanC[4];
	for (int i=0; i<4; i++)
		{
		kahanSum[i]=0.0;
		kahanC[i]=0.0;
		}
	double kahanY[4];
	double kahanT[4];

	double pertM[25][25];
	double iPertM[25][25];
	for (int k=0; k<25; k++)
		{
		for (int l=0; l<25; l++)
			{
			pertM[k][l]=0.0;
			iPertM[k][l]=0.0;
			}
		}

	double ranNb;
	double cPert[25];

	srand(time(NULL));
	for (int i=0; i<stat; i++)
		{

		for (int k=0; k<4; k++)
			{
			ranNb=((double)rand()/(double)RAND_MAX);
			for (int l=0; l<4; l++) pertM[k][l] = M[k][l] + eps*( 2.0*ranNb -1.0);
			}

		cofactor(pertM,dim,iPertM);
		prod(iPertM,y,cPert);

		for (int j=0; j<4; j++)
			{
			kahanY[j] = cPert[j] - kahanC[j];
			kahanT[j] = kahanSum[j] + kahanY[j];
			kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j];
			kahanSum[j] = kahanT[j];
			}
		}

	std::cout << "" << std::endl;
	std::cout << "Truth" << std::endl;
	for (int i=0; i<4; i++) std::cout << cfs[i] << "\t";
	std::cout << "" << std::endl;

	std::cout << "" << std::endl;
	std::cout << "AVERAGED" << std::endl;
	for (int i=0; i<4; i++) std::cout << kahanSum[i]/stat << "\t";
	std::cout << "" << std::endl;
	for (int i=0; i<4; i++) std::cout << abs(kahanSum[i]/stat - cfs[i]) << "\t";
	std::cout << "" << std::endl;

	std::cout << "" << std::endl;
	std::cout << "SINGLE" << std::endl;
	for (int i=0; i<4; i++) std::cout << cfs_single[i] << "\t";
	std::cout << "" << std::endl;
	for (int i=0; i<4; i++) std::cout << abs(cfs_single[i] - cfs[i]) << "\t";
	std::cout << "" << std::endl;

	std::cout << "" << std::endl;
	std::cout << "RATIOS" << std::endl;
	for (int i=0; i<4; i++) std::cout << abs(cfs_single[i] - cfs[i])/abs(kahanSum[i]/stat - cfs[i]) << "\t";

	return 0;
	}

double determinant(double a[25][25], double k)
{
  double s = 1, det = 0, b[25][25];
  int i, j, m, n, c;
  if (k == 1)
    {
     return (a[0][0]);
    }
  else
    {
     det = 0;
     for (c = 0; c < k; c++)
       {
        m = 0;
        n = 0;
        for (i = 0;i < k; i++)
          {
            for (j = 0 ;j < k; j++)
              {
                b[i][j] = 0;
                if (i != 0 && j != c)
                 {
                   b[m][n] = a[i][j];
                   if (n < (k - 2))
                    n++;
                   else
                    {
                     n = 0;
                     m++;
                     }
                   }
               }
             }
          det = det + s * (a[0][c] * determinant(b, k - 1));
          s = -1 * s;
          }
    }

    return (det);
}

void cofactor(double num[25][25], double f, double inverse[25][25])
{
 double b[25][25], fac[25][25];
 int p, q, m, n, i, j;
 for (q = 0;q < f; q++)
 {
   for (p = 0;p < f; p++)
    {
     m = 0;
     n = 0;
     for (i = 0;i < f; i++)
     {
       for (j = 0;j < f; j++)
        {
          if (i != q && j != p)
          {
            b[m][n] = num[i][j];
            if (n < (f - 2))
             n++;
            else
             {
               n = 0;
               m++;
               }
            }
        }
      }
      fac[q][p] = pow(-1, q + p) * determinant(b, f - 1);
    }
  }
  transpose(num, fac, f, inverse);
}
/*Finding transpose of matrix*/
void transpose(double num[25][25], double fac[25][25], double r, double inverse[25][25])
{
  int i, j;
  double b[25][25], d;

  for (i = 0;i < r; i++)
    {
     for (j = 0;j < r; j++)
       {
         b[i][j] = fac[j][i];
        }
    }
  d = determinant(num, r);
  for (i = 0;i < r; i++)
    {
     for (j = 0;j < r; j++)
       {
        inverse[i][j] = b[i][j] / d;
        }
    }
}
