/*
 * Simple_15.cpp
 *
 *  Created on: 17. 4. 2024
 *      Author: andrej
 */

#include <iostream>
#include <cstdlib>
#include <cmath>
#include "config.h"
using namespace std;

const int nHistoBins = 50;

double a(int n, double x)
	{
	if (n==1) return x;
	else if (n==2) return x*x;
	else return (2.0*x)*a(n-2,x)+(x-2.0)*a(n-1,x);
	}

static int binNb(double xMin, double xMax, int nBins, double x)
		{
		double dx = (xMax-xMin)/nBins;
		double delta = x-xMin;
		return (int)floor(delta/dx);
		}

static void createHisto(double histo[2][nHistoBins], double xMin,double xMax)
	{
	double dx = (xMax-xMin)/nHistoBins;
	histo[0][0] = xMin + dx/2.0;

	for (int i=0; i<nHistoBins; i++)
		{
		histo[0][i] = histo[0][0] + i*dx;
		histo[1][i] = 0.0;
		}
	}

static void fillHisto(double histo[2][nHistoBins], double x)
	{
	int L  = sizeof(histo[0]) / sizeof(histo[0][0]);
	double dx = (histo[0][L-1]-histo[0][0])/(L-1.0);
	double xMin = histo[0][0]-dx/2.0;
	double xMax = histo[0][L-1]+dx/2.0;
	int bin = binNb(xMin,xMax,L,x);
	if (0<=bin && bin<L) histo[1][bin]++;
	}

static void normHisto(double histo[2][nHistoBins])
	{
	int L  = sizeof(histo[0]) / sizeof(histo[0][0]);
	double sum = 0.0;
	for (int i=0; i<L; i++) sum = sum + histo[1][i];
	for (int i=0; i<L; i++) histo[1][i] = histo[1][i]/sum;
	}

int main(int argc, char **argv)
	{
	static int stat = (int)1E6;
	static double eps = 1E-7;
	static double x0 = 1.0/3.0;
	static int n = 15;
	static double width = 5E-13;

	double truth = 0.0000000696917193762563239137308507191523368295578192819843351;

	double ranNb;
	double x,y;
	double xMin = truth-width;
	double xMax = truth+width;

	double histo[2][nHistoBins];
	createHisto(histo, xMin, xMax);

	double kahanSum = 0.0;
	double kahanC = 0.0;

	srand(time(NULL));
	for (int i=0; i<stat; i++)
		{
		ranNb=((double)rand()/(double)RAND_MAX);
		x = x0 + eps*( 2.0*ranNb - 1.0 );
		y = a(n,x);

		double kahanY = y - kahanC;
		double kahanT = kahanSum + kahanY;
		kahanC = (kahanT - kahanSum) - kahanY;
		kahanSum = kahanT;

		fillHisto(histo,y);
		}

	double single = a(n,x0);
	double d_st = std::abs(single-truth);

	double avg = kahanSum/stat;
	double d_av = std::abs(avg-truth);

	double d_rat = d_st/d_av;
	std::cout.precision(17);
	std::cout << "Truth = " << truth <<std::endl;
	std::cout << "AV = " << avg << ", delta = " << d_av << std::endl;
	std::cout << "Single = " << single << ", delta = " << d_st << std::endl;
	std::cout << "Sing/AV delta  = " << d_rat <<std::endl;

	normHisto(histo);
	std::cout.precision(5);
	std::cout << "" << std::endl;
	for (int i=0; i<nHistoBins; i++)
		{
		double xVal = histo[0][i];
		double yVal = histo[1][i];
		std::cout <<  "(" << xVal << "," << yVal << ") "  << " ";
		}

	return 0;
	}
