/*
By: Andrej Liptaj
Date: 20. 6. 2024
*/
import java.util.Locale;
import org.apfloat.Apfloat;

import static java.lang.Math.floor;
import static org.apfloat.ApfloatMath.*;

public class D4_abrPrec
	{
	final static long prec = 300;
	final static int stat = 1000;
	final static double e0_denom = 1000;

	public static Apfloat fn(Apfloat x)
		{
		return exp(x).divide( pow(sin(x),3).add(pow(cos(x),3)) );
		}

	public static Apfloat xn(Apfloat x0, Apfloat e, Apfloat n)
		{
		return x0.add(e.multiply(n));
		}

	public static Apfloat D4(Apfloat x0, Apfloat e)
		{
		Apfloat y_m3 = xn(x0, e, new Apfloat(-3.0,prec));
		Apfloat y_m2 = xn(x0, e, new Apfloat(-2.0,prec));
		Apfloat y_m1 = xn(x0, e, new Apfloat(-1.0,prec));
		Apfloat y_0 = xn(x0, e, new Apfloat(0.0,prec));
		Apfloat y_p1 = xn(x0, e, new Apfloat(1.0,prec));
		Apfloat y_p2 = xn(x0, e, new Apfloat(2.0,prec));
		Apfloat y_p3 = xn(x0, e, new Apfloat(3.0,prec));

		y_m3 = fn(y_m3);
		y_m2 = fn(y_m2);
		y_m1 = fn(y_m1);
		y_0 = fn(y_0);
		y_p1 = fn(y_p1);
		y_p2 = fn(y_p2);
		y_p3 = fn(y_p3);

		y_m3 = y_m3.divide(new Apfloat(6.0,prec));
		y_m1 = y_m1.divide(new Apfloat(2.0,prec));
		y_0 = y_0.divide(new Apfloat(3.0,prec));
		y_p1 = y_p1.divide(new Apfloat(2.0,prec));
		y_p3 = y_p3.divide(new Apfloat(6.0,prec));

		y_m3 = y_m3.multiply(new Apfloat(-1.0,prec));
		y_m2 = y_m2.multiply(new Apfloat(2.0,prec));
		y_m1 = y_m1.multiply(new Apfloat(-13.0,prec));
		y_0 = y_0.multiply(new Apfloat(28.0,prec));
		y_p1 = y_p1.multiply(new Apfloat(-13.0,prec));
		y_p2 = y_p2.multiply(new Apfloat(2.0,prec));
		y_p3 = y_p3.multiply(new Apfloat(-1.0,prec));

		return (y_m3.add(y_m2).add(y_m1).add(y_0).add(y_p1).add(y_p2).add(y_p3)).divide(pow(e,4));
		}

	/* HISTOS */

	static int binNb(double xMin, double xMax, int nBins, double x)
		{
		double dx = (xMax-xMin)/nBins;
		double delta = x-xMin;
		return (int)floor(delta/dx);
		}

	static double[][] createHisto(double xMin,double xMax,int nBins)
		{
		double[][] histo = new double[2][nBins];
		double dx = (xMax-xMin)/nBins;
		histo[0][0] = xMin + dx/2.0;

		for (int i=0; i<nBins; i++)
			{
			histo[0][i] = histo[0][0] + i*dx;
			histo[1][i] = 0.0;
			}
		return histo;
		}

	static void fillHisto(double[][] histo, double x)
		{
		int L = histo[0].length;
		double dx = (histo[0][L-1]-histo[0][0])/(L-1.0);
		double xMin = histo[0][0]-dx/2.0;
		double xMax = histo[0][L-1]+dx/2.0;
		int bin = binNb(xMin,xMax,L,x);
		if (0<=bin && bin<L) histo[1][bin]++;
		//else System.out.println("Outlier!");
		}

	static void normHisto(double[][] histo)
		{
		int L = histo[0].length;
		double sum = 0.0;
		for (int i=0; i<L; i++) sum = sum + histo[1][i];

		if (sum==0.0)
			{
			System.out.println("Histogram cannot be normalized, is empty!");
			System.exit(-9);
			}

		for (int i=0; i<L; i++) histo[1][i] = histo[1][i]/sum;
		}

	static void printHisto(double[][] histo)
		{
		System.out.println();
		int L = histo[0].length;
		for (int i=0; i<L; i++)
			{
			String xVal = String.format("%.16f", (histo[0][i]) - 28.0);
			String yVal = String.format("%.16f", histo[1][i]);
			System.out.print( "(" + xVal + "," + yVal + ") " );
			}
		System.out.println();

		System.out.println();
		double sum = 0.0;
		double sumW = 0.0;
		for (int i=0; i<L; i++)
			{
			sum = sum + histo[0][i]*histo[1][i];
			sumW = sumW + histo[1][i];
			}
		System.out.println("Histo average is = " + (sum-28.0));
		System.out.println("Histo average is = " + sum/sumW);
		}

	public static void main(String[] args)
		{
		Locale.setDefault(new Locale("US") );
		Apfloat x0 = new Apfloat(0.0,prec);

		Apfloat em2 = new Apfloat(1.0,prec).divide(new Apfloat(10000.0,prec));
		Apfloat em1 = new Apfloat(1.0,prec).divide(new Apfloat(1000.0,prec));
		Apfloat e0 = new Apfloat(1.0,prec).divide(new Apfloat(e0_denom,prec));
		Apfloat ep1 = new Apfloat(1.0,prec).divide(new Apfloat(10.0,prec));
		Apfloat ep2 = new Apfloat(1.0,prec).divide(new Apfloat(1.0,prec));

		Apfloat truth = new Apfloat(28.0, prec);
/*
		System.out.println();
		Apfloat result = D4(x0,em2);
		System.out.println("D4 m2 = " + result);
		System.out.println("dis = " + result.subtract(truth) );

		System.out.println();
		result = D4(x0,em1);
		System.out.println("D4 m1 = " + result);
		System.out.println("dis = " + result.subtract(truth) );

		System.out.println();
		result = D4(x0,e0);
		System.out.println("D4 0 = " + result);
		System.out.println("dis = " + result.subtract(truth) );

		System.out.println();
		result = D4(x0,ep1);
		System.out.println("D4 p1 = " + result);
		System.out.println("dis = " + result.subtract(truth) );

		System.out.println();
		result = D4(x0,ep1);
		System.out.println("D4 p2 = " + result);
		System.out.println("dis = " + result.subtract(truth) );
*/
		/***************/

		Apfloat de = e0.divide( new Apfloat(stat,prec) );
		Apfloat e =  e0.divide(new Apfloat(2.0,prec));
		Apfloat av = new Apfloat(0.0, prec);
		Apfloat[] memory = new Apfloat[stat];
		int greater = 0;
		int smaller = 0;
		int e_greater = 0;
		int e_smaller = 0;

		for (int i=1; i<stat+1; i++)
			{
			Apfloat val = D4(x0,e);
			memory[i-1] = val;

			av = av.add( (val.subtract(av)).divide( new Apfloat((float)i,prec) )  );

			int isGrater = val.compareTo( truth );
			if (isGrater==1) greater++;
				else smaller++;

			int isGrater_e = e.compareTo( e0 );
			if (isGrater_e==1) e_greater++;
			else e_smaller++;

			e = e.add(de);
			}

		System.out.println();
		System.out.println();
		System.out.println("AV = " + av);
		System.out.println("greater = " + greater);
		System.out.println("smaller = " + smaller);
		System.out.println("e_greater = " + e_greater);
		System.out.println("e_smaller = " + e_smaller);

		System.out.println();
		System.out.println("Histos");
		System.out.println();

		double[][] histo = createHisto(28.0-1E-8,28.0+1E-8,50);
		for (int i=0; i<stat;i++)
			{
			double dVal = memory[i].doubleValue();
			fillHisto(histo,dVal);
			}

		normHisto(histo);
		printHisto(histo);
		System.out.println("precision = " + prec);
		System.out.println("statistics = " + stat);
		}

	}
