/*
By: Andrej Liptaj
Date: 15. 4. 2024
*/

import java.security.SecureRandom;
import java.util.Locale;
import java.util.Random;

import static java.lang.Math.*;
import static java.lang.Math.cos;

public class Simple_f4
	{
	static final double stat = 1E7;

	static final double dx = 1E-14;
	static final double h0_d1 = 1E-4;
	static final double h0_d2 = 1E-3;
	static final double h0_d3 = 1E-3;

	//static final double h0_d4 = 1E-2;
	static final double h0_d4 = 1E-2/2;
	static final double width_d4 = 1.4E-5;

	static final boolean h_isDisturbed=true;
	static final double x0 = 0;

	static final double truth_d1 = 1;
	static final double truth_d2 = 4;
	static final double truth_d3 = 4;
	static final double truth_d4 = 28;


	static final int nHistoBins = 50;


	static double fun(double x)
		{
		return exp(x)/( pow(sin(x),3.0) + pow(cos(x),3.0) );
		}

	static double d1(double h, double x0)
		{
		double[] x = {x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h};
		double[] y = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]) };
		return (y[0]/12.0-2.0*y[1]/3.0+2.0*y[3]/3.0-y[4]/12.0)/pow(h,1);
		}

	static double d2(double h, double x0)
		{
		double[] x = {x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h};
		double[] y = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]) };
		return (-y[0]/12.0+4.0*y[1]/3.0-5.0*y[2]/2.0+4.0*y[3]/3.0-y[4]/12.0)/pow(h,2);
		}

	static double d3(double h, double x0)
		{
		double[] x = {x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h};
		double[] y = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6]) };
		return (y[0]/8.0-y[1]+13.0*y[2]/8.0-13.0*y[4]/8.0+y[5]-y[6]/8.0)/pow(h,3);
		}

	static double d4(double h, double x0)
		{
		double[] x = {x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h};
		double[] y = { fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6]) };
		return (-y[0]/6.0+2.0*y[1]-13.0*y[2]/2.0+28.0*y[3]/3.0-13.0*y[4]/2.0+2.0*y[5]-y[6]/6.0)/pow(h,4);
		}

	/* HISTOS */

	static int binNb(double xMin, double xMax, int nBins, double x)
		{
		double dx = (xMax-xMin)/nBins;
		double delta = x-xMin;
		return (int)floor(delta/dx);
		}

	static double[][] createHisto(double xMin,double xMax,int nBins)
		{
		double[][] histo = new double[2][nBins];
		double dx = (xMax-xMin)/nBins;
		histo[0][0] = xMin + dx/2.0;

		for (int i=0; i<nBins; i++)
			{
			histo[0][i] = histo[0][0] + i*dx;
			histo[1][i] = 0.0;
			}
		return histo;
		}

	static void fillHisto(double[][] histo, double x)
		{
		int L = histo[0].length;
		double dx = (histo[0][L-1]-histo[0][0])/(L-1.0);
		double xMin = histo[0][0]-dx/2.0;
		double xMax = histo[0][L-1]+dx/2.0;
		int bin = binNb(xMin,xMax,L,x);
		if (0<=bin && bin<L) histo[1][bin]++;
		//else System.out.println("Outlier!");
		}

	static void normHisto(double[][] histo)
		{
		int L = histo[0].length;
		double sum = 0.0;
		for (int i=0; i<L; i++) sum = sum + histo[1][i];

		if (sum==0.0)
			{
			System.out.println("Histogram cannot be normalized, is empty!");
			System.exit(-9);
			}

		for (int i=0; i<L; i++) histo[1][i] = histo[1][i]/sum;
		}

	static void printHisto(double[][] histo)
		{
		System.out.println();
		int L = histo[0].length;
		for (int i=0; i<L; i++)
			{
			String xVal = String.format("%.7f", histo[0][i]-28.0);

			String yVal = String.format("%.7f", histo[1][i]);
			System.out.print( "(" + xVal + "," + yVal + ") " );
			//System.out.print( "(" + i + "," + yVal + ") " );
			}
		System.out.println();

		System.out.println();
		double sum = 0.0;
		double sumW = 0.0;
		for (int i=0; i<L; i++)
			{
			sum = sum + histo[0][i]*histo[1][i];
			sumW = sumW + histo[1][i];
			}
		System.out.println("Histo average is = " + (sum-28.0));
		System.out.println("Histo average is = " + sum/sumW);
		}

	public static void main(String[] args)
		{
		Locale.setDefault(new Locale("US") );

		Random rnd = new SecureRandom();
		double[] kahanSum = new double[4];
		double[] kahanC = new double[4];
		for (int i=0; i<4; i++)
			{
			kahanSum[i] = 0.0;
			kahanC[i] = 0.0;
			}

		//double[][] histo_d4 = createHisto(truth_d4-width_d4,truth_d4,nHistoBins);
		//double[][] histo_d4 = createHisto(27.791928854240897-1E-10,27.791928854240897+1E-10,nHistoBins);
		//double[][] histo_d4 = createHisto(truth_d4-width_d4/100.0,truth_d4+width_d4/50.0,nHistoBins);
		double[][] histo_d4 = createHisto(truth_d4-width_d4,truth_d4+width_d4,nHistoBins);

		for (int i=0; i<stat; i++)
			{
			double x = x0;
			double h;
			double[] di = new double[4];

			if (h_isDisturbed)
				{
				h = h0_d1 *( rnd.nextDouble()+0.5 );
				di[0] = d1(h,x);

				h = h0_d2 *( rnd.nextDouble()+0.5 );
				di[1] = d2(h,x);

				h = h0_d3 *( rnd.nextDouble()+0.5 );
				di[2] = d3(h,x);

				//h = h0_d4 *( rnd.nextDouble()+0.5 );
				h = h0_d4 + 1E-10*( 2.0*rnd.nextDouble()-1.0 );
				di[3] = d4(h,x);

				double[] kahanY = new double[4];
				double[] kahanT = new double[4];
				for (int j=0; j<4; j++)
					{
					kahanY[j] = di[j] - kahanC[j];
					kahanT[j] = kahanSum[j] + kahanY[j];
					kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j];
					kahanSum[j] = kahanT[j];
					}

				fillHisto(histo_d4,di[3]);
				}
			else
				{
				x = x0 + dx*( 2.0*rnd.nextDouble() -1.0);

				di[0] = d1(1E-4,x);
				di[1] = d2(1E-3,x);
				di[2] = d3(1E-3,x);
				di[3] = d4(1E-2,x);

				double[] kahanY = new double[4];
				double[] kahanT = new double[4];
				for (int j=0; j<4; j++)
					{
					kahanY[j] = di[j] - kahanC[j];
					kahanT[j] = kahanSum[j] + kahanY[j];
					kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j];
					kahanSum[j] = kahanT[j];
					}


				fillHisto(histo_d4,di[3]);
				}

			}

		double df1_si = d1(1E-4,x0);
		double d1_si = abs(truth_d1 -df1_si);
		double df1_av = kahanSum[0]/stat;
		double d1_av = abs(truth_d1 -df1_av);
		
		double df2_si = d2(1E-3,x0);
		double d2_si = abs(truth_d2 -df2_si);
		double df2_av = kahanSum[1]/stat;
		double d2_av = abs(truth_d2 -df2_av);

		double df3_si = d3(1E-3,x0);
		double d3_si = abs(truth_d3 -df3_si);
		double df3_av = kahanSum[2]/stat;
		double d3_av = abs(truth_d3 -df3_av);

		double df4_si = d4(1E-2,x0);
		double d4_si = abs(truth_d4 -df4_si);
		double df4_av = kahanSum[3]/stat;
		double d4_av = abs(truth_d4 -df4_av);

		System.out.println();
		System.out.println("First Derivative");
		System.out.println("Truth = " + truth_d1);
		System.out.println("AV = " + df1_av + ", delta = " +  d1_av );
		System.out.println("Single = " + df1_si + ", delta = " + d1_si );
		System.out.println("Sing/AV delta  = " + d1_si/d1_av );

		System.out.println();
		System.out.println("Second Derivative");
		System.out.println("Truth = " + truth_d2);
		System.out.println("AV = " + df2_av + ", delta = " +  d2_av );
		System.out.println("Single = " + df2_si + ", delta = " + d2_si );
		System.out.println("Sing/AV delta  = " + d2_si/d2_av );

		System.out.println();
		System.out.println("Third Derivative");
		System.out.println("Truth = " + truth_d3);
		System.out.println("AV = " + df3_av + ", delta = " +  d3_av );
		System.out.println("Single = " + df3_si + ", delta = " + d3_si );
		System.out.println("Sing/AV delta  = " + d3_si/d3_av );

		System.out.println();
		System.out.println("Fourth Derivative");
		System.out.println("Truth = " + truth_d4);
		System.out.println("AV = " + df4_av + ", delta = " +  d4_av );
		System.out.println("Single = " + df4_si + ", delta = " + d4_si );
		System.out.println("Sing/AV delta  = " + d4_si/d4_av );

		/*HISTO*/

		System.out.println();
		System.out.println("HISTO");
		System.out.println();
		normHisto(histo_d4);
		printHisto(histo_d4);

		double mod_d4_truth = 0.0;
		System.out.println("mod_d4_truth = " + mod_d4_truth);
		System.out.println("single = " + (df4_si-28.0));
		//System.out.println("mod_d4_single = " + mod_d4_single);
		//System.out.println("mod_d4_avg = " + mod_d4_avg);


		}
	}
