/*
By: Andrej Liptaj
Date: 16. 4. 2024
*/

import java.security.SecureRandom;
import java.util.Locale;
import java.util.Random;

import static java.lang.Math.*;

public class Simple_r13
	{
	final static int stat = (int)1E6;
	final static double eps = 1E-4;
	final static double x0 = 10.0;
	final static int n = 13;
	final static double truth = 0.007878519876053262;
	final static double width = 7E-3;
	final static int nHistoBins = 50;

	static double rn(int n, double x)
		{
		if (n==0) return log(x)-log(x-1.0);
		else return x*rn(n-1,x)-1.0/n;
		}

	static int binNb(double xMin, double xMax, int nBins, double x)
		{
		double dx = (xMax-xMin)/nBins;
		double delta = x-xMin;
		return (int)floor(delta/dx);
		}

	static double[][] createHisto(double xMin,double xMax,int nBins)
		{
		double[][] histo = new double[2][nBins];
		double dx = (xMax-xMin)/nBins;
		histo[0][0] = xMin + dx/2.0;

		for (int i=0; i<nBins; i++)
			{
			histo[0][i] = histo[0][0] + i*dx;
			histo[1][i] = 0.0;
			}
		return histo;
		}

	static void fillHisto(double[][] histo, double x)
		{
		int L = histo[0].length;
		double dx = (histo[0][L-1]-histo[0][0])/(L-1.0);
		double xMin = histo[0][0]-dx/2.0;
		double xMax = histo[0][L-1]+dx/2.0;
		int bin = binNb(xMin,xMax,L,x);
		if (0<=bin && bin<L) histo[1][bin]++;
			//else System.out.println("Outlier!");
		}

	static void normHisto(double[][] histo)
		{
		int L = histo[0].length;
		double sum = 0.0;
		for (int i=0; i<L; i++) sum = sum + histo[1][i];

		if (sum==0.0)
			{
			System.out.println("Histogram cannot be normalized, is empty!");
			System.exit(-9);
			}

		for (int i=0; i<L; i++) histo[1][i] = histo[1][i]/sum;
		}

	static void printHisto(double[][] histo)
		{
		System.out.println();
		int L = histo[0].length;
		for (int i=0; i<L; i++)
			{
			String xVal = String.format("%.5f", histo[0][i]);
			String yVal = String.format("%.5f", histo[1][i]);
			System.out.print( "(" + xVal + "," + yVal + ") " );
			}
		System.out.println();

		System.out.println();
		double sum = 0.0;
		for (int i=0; i<L; i++)
			{
			sum = sum + histo[0][i]*histo[1][i];
			}
		System.out.println("Histo average is = " + sum);
		}

	public static void main(String[] args)
		{
		Locale.setDefault(new Locale("US") );

		Random rng = new SecureRandom();
		double kahanSum = 0.0;
		double kahanC = 0.0;

		double flyingAv = 0.0;
		double xMin = truth-width;
		double xMax = truth+width;
		double[][] histo = createHisto(xMin,xMax,nHistoBins);

		for (int i=0; i<stat; i++)
			{
			double x = x0 + eps*( 2.0*rng.nextDouble() - 1.0 );
			double y = rn(n,x);

			double kahanY = y - kahanC;
			double kahanT = kahanSum + kahanY;
			kahanC = (kahanT-kahanSum) - kahanY;
			kahanSum = kahanT;

			fillHisto(histo,y);

			flyingAv = (i*flyingAv + y)/(i+1);
			}

		double single = rn(n,x0);
		double d_st = abs(single-truth);

		double avg = kahanSum/stat;
		double d_at = abs(avg-truth);

		double d_rat = d_st/d_at;
		System.out.println("Truth = " + truth);
		System.out.println("AV = " + avg + ", delta = " +  d_at);
		System.out.println("Single = " + single+ ", delta = " + d_st);
		System.out.println("Sing/AV delta  = " + d_rat);
		System.out.println("Ceck: AV-flyAV  = " + (avg-flyingAv) );
		System.out.println("Ceck: AV-Single  = " + (avg-single) );

		normHisto(histo);
		printHisto(histo);
		}

	}
