"use strict";

const stat = 1E5;

const dx = 1E-15;

const h0_d1 = 1E-4;
const h0_d2 = 1E-3;
const h0_d3 = 1E-3;
const h0_d4 = 1E-2;

const x0 = 0;

const truth_d1 = 1;
const truth_d2 = 4;
const truth_d3 = 4;
const truth_d4 = 28;

const h_isDisturbed=false;

let truth = 4;
const width = 0.2E-9;
const nHistoBins = 50;

alert("Confirm program start!");
document.writeln("STARTED!" +  "<br>");

function fun(x)
	{
	return Math.exp(x)/( Math.pow(Math.sin(x),3.0) + Math.pow(Math.cos(x), 3.0) );
	}

function d1(h, x0)
	{
	const x = [x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h];
	const y = [ fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]) ];
	return (y[0]/12.0-2.0*y[1]/3.0+2.0*y[3]/3.0-y[4]/12.0)/Math.pow(h,1);
	}

function d2(h, x0)
	{
    const x = [x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h];
	const y = [ fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]) ];
	return (-y[0]/12.0+4.0*y[1]/3.0-5.0*y[2]/2.0+4.0*y[3]/3.0-y[4]/12.0)/Math.pow(h,2);
	}

function d3(h, x0)
	{
    const x = [x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h];
	const y = [fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6]) ];
	return (y[0]/8.0-y[1]+13.0*y[2]/8.0-13.0*y[4]/8.0+y[5]-y[6]/8.0)/Math.pow(h,3);
	}

function d4(h, x0)
	{
    const x = [x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h];
	const y = [ fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6]) ];
	return (-y[0]/6.0+2.0*y[1]-13.0*y[2]/2.0+28.0*y[3]/3.0-13.0*y[4]/2.0+2.0*y[5]-y[6]/6.0)/Math.pow(h,4);
	}

function binNb(xMin,xMax,nBins,x)
    {
    let dx = (xMax-xMin)/nBins;
    let delta = x-xMin;
    return Math.floor(delta/dx);
    }

function createHisto(xMin, xMax, nBins)
    {
    let histo = [2];
    histo[0] = [nBins];
    histo[1] = [nBins];

    let dx = (xMax-xMin)/nBins;
    histo[0][0] = xMin + dx/2.0;

    for (let i=0; i<nBins; i++)
        {
        histo[0][i] = histo[0][0] + i*dx;
        histo[1][i] = 0.0;
        }
    return histo;
    }


function fillHisto(histo, x)
    {
    let L = histo[0].length;
    let dx = (histo[0][L-1]-histo[0][0])/(L-1.0);
    let xMin = histo[0][0]-dx/2.0;
    let xMax = histo[0][L-1]+dx/2.0;
    let bin = binNb(xMin,xMax,L,x);
    if (0<=bin && bin<L) histo[1][bin]++;
    }

function normHisto(histo)
    {
    let L = histo[0].length;
    let sum = 0.0;
    for (let i=0; i<L; i++) sum = sum + histo[1][i];
    for (let i=0; i<L; i++) histo[1][i] = histo[1][i]/sum;
    }

 let kahanSum = [4];
 let kahanC = [4];
 for (let i=0; i<4; i++)
     {
     kahanSum[i]=0.0;
     kahanC[i]=0.0;
     }

let xMin = truth-width;
let xMax = truth+width;
let histo = createHisto(xMin,xMax,nHistoBins);

for (let i=0; i<stat; i++)
     {
     let x = x0;
     let h;
     let di = [4];

     if (h_isDisturbed)
        {
        h = h0_d1 *( Math.random()+0.5 );
        di[0] = d1(h,x);
        
        h = h0_d2 *( Math.random()+0.5 );
        di[1] = d2(h,x);
        
        h = h0_d3 *( Math.random()+0.5 );
        di[2] = d3(h,x);

        h = h0_d4 *( Math.random()+0.5 );
        di[3] = d4(h,x);

        let kahanY = [4];
        let kahanT = [4];
        for (let j=0; j<4; j++)
            {
            kahanY[j] = di[j] - kahanC[j];
            kahanT[j] = kahanSum[j] + kahanY[j];
            kahanC[j] = (kahanT[j] - kahanSum[j] ) - kahanY[j];
            kahanSum[j] = kahanT[j];
            }
        
        fillHisto(histo,di[1]);
        }
    else
        {
        x = x0 + dx*( 2.0*Math.random() -1.0);

        di[0] = d1(1E-4,x);
        di[1] = d2(1E-3,x);
        di[2] = d3(1E-3,x);
        di[3] = d4(1E-2,x);

        let kahanY = [4];
        let kahanT = [4];
	    for (let j=0; j<4; j++)
            {
            kahanY[j] = di[j] - kahanC[j];
            kahanT[j] = kahanSum[j] + kahanY[j];
            kahanC[j] = (kahanT[j] - kahanSum[j] ) - kahanY[j];
            kahanSum[j] = kahanT[j];
            }
        
        fillHisto(histo,di[1]);

        }
    }

let df1_si = d1(1E-4,x0);
let d1_si = Math.abs(truth_d1 -df1_si);
let df1_av = kahanSum[0]/stat;
let d1_av = Math.abs(truth_d1 -df1_av);
    
let df2_si = d2(1E-3,x0);
let d2_si = Math.abs(truth_d2 -df2_si);
let df2_av = kahanSum[1]/stat;
let d2_av = Math.abs(truth_d2 -df2_av);

let df3_si = d3(1E-3,x0);
let d3_si = Math.abs(truth_d3 -df3_si);
let df3_av = kahanSum[2]/stat;
let d3_av = Math.abs(truth_d3 -df3_av);

let df4_si = d4(1E-2,x0);
let d4_si = Math.abs(truth_d4 -df4_si);
let df4_av = kahanSum[3]/stat;
let d4_av = Math.abs(truth_d4 -df4_av);

document.writeln("<br>");
document.writeln("First Derivative"+ "<br>");
document.writeln("Truth = " + truth_d1+ "<br>");
document.writeln("AV = " + df1_av + ", delta = " +  d1_av + "<br>");
document.writeln("Single = " + df1_si + ", delta = " + d1_si + "<br>");
document.writeln("Sing/AV delta  = " + d1_si/d1_av + "<br>");
    
document.writeln("<br>");
document.writeln("Second Derivative"+ "<br>");
document.writeln("Truth = " + truth_d2+ "<br>");
document.writeln("AV = " + df2_av + ", delta = " +  d2_av + "<br>");
document.writeln("Single = " + df2_si + ", delta = " + d2_si + "<br>");
document.writeln("Sing/AV delta  = " + d2_si/d2_av + "<br>");

document.writeln("<br>");
document.writeln("Third Derivative"+ "<br>");
document.writeln("Truth = " + truth_d3+ "<br>");
document.writeln("AV = " + df3_av + ", delta = " +  d3_av + "<br>");
document.writeln("Single = " + df3_si + ", delta = " + d3_si + "<br>");
document.writeln("Sing/AV delta  = " + d3_si/d3_av + "<br>");

document.writeln("<br>");
document.writeln("Fourth Derivative"+ "<br>");
document.writeln("Truth = " + truth_d4+ "<br>");
document.writeln("AV = " + df4_av + ", delta = " +  d4_av + "<br>");
document.writeln("Single = " + df4_si + ", delta = " + d4_si + "<br>");
document.writeln("Sing/AV delta  = " + d4_si/d4_av + "<br>");

normHisto(histo);
document.writeln("<br>");
for (let i=0; i<nHistoBins; i++)
	{
	let xVal = 1E0*(histo[0][i]-4);
	let yVal = histo[1][i];
	document.writeln( "(" + xVal.toFixed(13) + "," + yVal.toFixed(10) + ") " );
	}