import math
import numpy as np
import random

stat = int(1E5)
nHistoBins = 50
eps = 1E-14
cfs = [1.0, -1.0 / 3.0, 1.0 / 7.0, -1.0 / 11.0]


truth_c0 = 1
truth_c1 = -1/3
truth_c2 = 1/7
truth_c3 = -1/11

# width_c0 = 0.003
width_c0 = 0.03
width_c1 = 0.1
width_c2 = 0.1
width_c3 = 0.1


def binNb(xMin, xMax, nBins, x):
    dx = (xMax-xMin)/nBins
    delta = x-xMin
    return int(math.floor(delta/dx))


def createHisto(xMin, xMax, nBins):
    histo = np.zeros([2,nBins])
    dx = (xMax-xMin)/nBins
    histo[0][0] = xMin + dx/2.0
    for i in range(0, nBins):
        histo[0][i] = histo[0][0] + i*dx
        histo[1][i] = 0.0
    return histo


def fillHisto(histo, x):
    L = histo[0].size
    dx = (histo[0][L-1]-histo[0][0])/(L-1.0)
    xMin = histo[0][0]-dx/2.0
    xMax = histo[0][L-1]+dx/2.0
    bin = binNb(xMin,xMax,L,x)
    if (0<=bin & bin<L):
        histo[1][bin]=histo[1][bin]+1


def normHisto(histo):
    L = histo[0].size
    sum = 0.0
    for i in range(0,L):
        sum = sum + histo[1][i]
    for i in range(0,L):
        histo[1][i] = histo[1][i]/sum

sum_c0 = 0.0
sum_c1 = 0.0
sum_c2 = 0.0
sum_c3 = 0.0

xMin_c0 = truth_c0-width_c0
xMin_c1 = truth_c1-width_c1
xMin_c2 = truth_c2-width_c2
xMin_c3 = truth_c3-width_c3

xMax_c0 = truth_c0+width_c0
xMax_c1 = truth_c1+width_c1
xMax_c2 = truth_c2+width_c2
xMax_c3 = truth_c3+width_c3

histo_c0 = createHisto(xMin_c0,xMax_c0,nHistoBins)
histo_c1 = createHisto(xMin_c1,xMax_c1,nHistoBins)
histo_c2 = createHisto(xMin_c2,xMax_c2,nHistoBins)
histo_c3 = createHisto(xMin_c3,xMax_c3,nHistoBins)


def fn(z):
    return cfs[0] + cfs[1] * math.cos(z) + cfs[2] * math.cos(2.0 * z) + cfs[3] * math.cos(3.0 * z)


d = 1E-4
x = [1.0, 1.0+d, 1.0+d+d, 1.0+d+d+d]
y = np.zeros([4])

for i in range(0, 4):
    y[i] = fn(x[i])

M = np.zeros([4, 4])
for i in range(0, 4):
    M[i][0] = 1.0
    M[i][1] = math.cos(x[i])
    M[i][2] = math.cos(2.0*x[i])
    M[i][3] = math.cos(3.0*x[i])


iMat = np.linalg.inv(M)
yVec = np.array([y[0], y[1], y[2], y[3]])

cfsVec_single = iMat@yVec

kahanSum = np.zeros([4])
kahanC = np.zeros([4])
kahanY = np.zeros([4])
kahanT = np.zeros([4])

for i in range(0, stat):
    pertM = np.zeros([4, 4])

    for k in range(0, 4):
        for m in range(0, 4):
            pertM[k][m] = M[k][m]+eps*(2.0*random.random() - 1.0)

    iPertM = np.linalg.inv(pertM)
    cPert = iPertM@yVec

    for j in range(0, 4):
        kahanY[j] = cPert[j] - kahanC[j]
        kahanT[j] = kahanSum[j] + kahanY[j]
        kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j]
        kahanSum[j] = kahanT[j]

    fillHisto(histo_c0,cPert[0])
    fillHisto(histo_c1,cPert[1])
    fillHisto(histo_c2,cPert[2])
    fillHisto(histo_c3,cPert[3])


print("Truth")
for i in range(0, 4):
    print(str(cfs[i]) + "\t", end="")
print()

print()
print("AVERAGED")
for i in range(0, 4):
    print(str(kahanSum[i] / stat) + "\t", end="")
print()
for i in range(0, 4):
    print(str(abs(kahanSum[i] / stat - cfs[i])) + "\t", end="")
print()

print()
print("SINGLE")
for i in range(0, 4):
    print(str(cfsVec_single[i]) + "\t", end="")
print()
for i in range(0, 4):
    print(str(abs(cfsVec_single[i] - cfs[i])) + "\t", end="")
print()

print()
print("RATIOS\n")
for i in range(0, 4):
    print(str(abs(cfsVec_single[i] - cfs[i]) / abs(kahanSum[i] / stat - cfs[i])) + "\t", end="")

print("\nHISTOS")
print()

normHisto(histo_c0)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_c0[0][i]-truth_c0)
    yVal = str(histo_c0[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")

normHisto(histo_c1)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_c1[0][i])
    yVal = str(histo_c1[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")

normHisto(histo_c2)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_c2[0][i])
    yVal = str(histo_c2[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")

normHisto(histo_c3)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_c3[0][i])
    yVal = str(histo_c3[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")