import math
import random
import numpy as np

stat = int(1E5)
h_isDisturbed = True

dx = 1E-6

h0_d1 = 1E-4
h0_d2 = 1E-3
h0_d3 = 1E-3
h0_d4 = 1E-2

nHistoBins = 50
x0 = 0

truth_d1 = 1
truth_d2 = 4
truth_d3 = 4
truth_d4 = 28

width_d1 = 0.1
width_d2 = 0.1
width_d3 = 0.1
width_d4 = 0.1


def fun(x):
    return math.exp(x)/(pow(math.sin(x),3.0) + pow(math.cos(x),3.0))


def d1(h, x0):
    x = [x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h]
    y = [fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4])]
    return (y[0]/12.0-2.0*y[1]/3.0+2.0*y[3]/3.0-y[4]/12.0)/pow(h, 1)


def d2(h, x0):
    x = [x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h]
    y = [fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4])]
    return (-y[0]/12.0+4.0*y[1]/3.0-5.0*y[2]/2.0+4.0*y[3]/3.0-y[4]/12.0)/pow(h, 2)


def d3(h, x0):
    x = [x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h]
    y = [fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6])]
    return (y[0]/8.0-y[1]+13.0*y[2]/8.0-13.0*y[4]/8.0+y[5]-y[6]/8.0)/pow(h, 3)


def d4(h, x0):
    x = [x0-3.0*h, x0-2.0*h, x0-h, x0, x0+h, x0+2.0*h, x0+3.0*h]
    y = [fun(x[0]), fun(x[1]), fun(x[2]), fun(x[3]), fun(x[4]), fun(x[5]), fun(x[6])]
    return (-y[0]/6.0+2.0*y[1]-13.0*y[2]/2.0+28.0*y[3]/3.0-13.0*y[4]/2.0+2.0*y[5]-y[6]/6.0)/pow(h, 4)


def binNb(xMin, xMax, nBins, x):
    dx = (xMax-xMin)/nBins
    delta = x-xMin
    return int(math.floor(delta/dx))

def createHisto(xMin, xMax, nBins):
    histo = np.zeros([2,nBins])
    dx = (xMax-xMin)/nBins
    histo[0][0] = xMin + dx/2.0
    for i in range(0, nBins):
        histo[0][i] = histo[0][0] + i*dx
        histo[1][i] = 0.0
    return histo

def fillHisto(histo, x):
    L = histo[0].size
    dx = (histo[0][L-1]-histo[0][0])/(L-1.0)
    xMin = histo[0][0]-dx/2.0
    xMax = histo[0][L-1]+dx/2.0
    bin = binNb(xMin,xMax,L,x)
    if (0<=bin & bin<L):
        histo[1][bin]=histo[1][bin]+1

def normHisto(histo):
    L = histo[0].size
    sum = 0.0
    for i in range(0,L):
        sum = sum + histo[1][i]
    for i in range(0,L):
        histo[1][i] = histo[1][i]/sum

xMin_d1 = truth_d1-width_d1
xMin_d2 = truth_d2-width_d2
xMin_d3 = truth_d3-width_d3
xMin_d4 = truth_d4-width_d4

xMax_d1 = truth_d1+width_d1
xMax_d2 = truth_d2+width_d2
xMax_d3 = truth_d3+width_d3
xMax_d4 = truth_d4+width_d4

histo_d1 = createHisto(xMin_d1,xMax_d1,nHistoBins)
histo_d2 = createHisto(xMin_d2,xMax_d2,nHistoBins)
histo_d3 = createHisto(xMin_d3,xMax_d3,nHistoBins)
histo_d4 = createHisto(xMin_d4,xMax_d4,nHistoBins)

kahanSum = np.zeros([4])
kahanC = np.zeros([4])
kahanY = np.zeros([4])
kahanT = np.zeros([4])
di = np.zeros([4])

for i in range(0, stat):
    x = x0

    if (h_isDisturbed):
        h = h0_d1 *( random.random()+0.5 )
        di[0] = d1(h,x)

        h = h0_d2 *( random.random()+0.5 )
        di[1] = d2(h,x)

        h = h0_d3 *( random.random()+0.5 )
        di[2] = d3(h,x)

        h = h0_d4 *( random.random()+0.5 )
        di[3] = d4(h,x)

        for j in range(0, 4):
            kahanY[j] = di[j] - kahanC[j]
            kahanT[j] = kahanSum[j] + kahanY[j]
            kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j]
            kahanSum[j] = kahanT[j]

        fillHisto(histo_d1,di[0])
        fillHisto(histo_d2,di[1])
        fillHisto(histo_d3,di[2])
        fillHisto(histo_d4,di[3])
    
    else:
        x = x0 + dx*( 2.0*random.random() - 1.0)

        di[0] = d1(1E-4,x)
        di[1] = d2(1E-3,x)
        di[2] = d3(1E-3,x)
        di[3] = d4(1E-2,x)

        for j in range(0, 4):
            kahanY[j] = di[j] - kahanC[j]
            kahanT[j] = kahanSum[j] + kahanY[j]
            kahanC[j] = (kahanT[j] - kahanSum[j]) - kahanY[j]
            kahanSum[j] = kahanT[j]

        fillHisto(histo_d1,di[0])
        fillHisto(histo_d2,di[1])
        fillHisto(histo_d3,di[2])
        fillHisto(histo_d4,di[3])

df1_si = d1(1E-4,x0)
d1_si = abs(truth_d1 -df1_si)
df1_av = kahanSum[0]/stat
d1_av = abs(truth_d1 -df1_av)

df2_si = d2(1E-3,x0)
d2_si = abs(truth_d2 -df2_si)
df2_av = kahanSum[1]/stat
d2_av = abs(truth_d2 -df2_av)

df3_si = d3(1E-3,x0)
d3_si = abs(truth_d3 -df3_si)
df3_av = kahanSum[2]/stat
d3_av = abs(truth_d3 -df3_av)

df4_si = d4(1E-2,x0)
d4_si = abs(truth_d4 -df4_si)
df4_av = kahanSum[3]/stat
d4_av = abs(truth_d4 -df4_av)

print()
print("First Derivative")
print("Truth = " + str(truth_d1))
print("AV = " + str(df1_av) + ", delta = " + str(d1_av) )
print("Single = " + str(df1_si) + ", delta = " + str(d1_si) )
if d1_av != 0.0:
    print("Sing/AV delta  = " + str(d1_si/d1_av) )
else:
    print("Sing/AV delta  = infinity")

print()
print("Second Derivative")
print("Truth = " + str(truth_d2))
print("AV = " + str(df2_av) + ", delta = " + str(d2_av) )
print("Single = " + str(df2_si) + ", delta = " + str(d2_si) )
print("Sing/AV delta  = " + str(d2_si/d2_av) )

print()
print("Third Derivative")
print("Truth = " + str(truth_d3))
print("AV = " + str(df3_av) + ", delta = " + str(d3_av) )
print("Single = " + str(df3_si) + ", delta = " + str(d3_si) )
print("Sing/AV delta  = " + str(d3_si/d3_av) )

print()
print("Fourth Derivative")
print("Truth = " + str(truth_d4))
print("AV = " + str(df4_av) + ", delta = " + str(d4_av) )
print("Single = " + str(df4_si) + ", delta = " + str(d4_si) )
print("Sing/AV delta  = " + str(d4_si/d4_av) )

"""
normHisto(histo_d1)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_d1[0][i])
    yVal = str(histo_d1[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")

normHisto(histo_d2)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_d2[0][i])
    yVal = str(histo_d2[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")

normHisto(histo_d3)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_d3[0][i])
    yVal = str(histo_d3[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")

normHisto(histo_d4)
print()
print()
for i in range (0, nHistoBins):
    xVal = str(histo_d4[0][i])
    yVal = str(histo_d4[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")
"""