import math
import random
import numpy as np

eps = 1E-4
x0 = 10.0
n = 13
stat = int(1E5)
nHistoBins = 50
width = 7E-3
truth = 0.007878519876053262

def rn(m, x):
    if m == 0:
        return math.log(x)-math.log(x-1.0)
    else:
        return x*rn(m-1, x)-1.0/m


def binNb(xMin, xMax, nBins, x):
    dx = (xMax-xMin)/nBins
    delta = x-xMin
    return int(math.floor(delta/dx))

def createHisto(xMin, xMax, nBins):
    histo = np.zeros([2,nBins])
    dx = (xMax-xMin)/nBins
    histo[0][0] = xMin + dx/2.0
    for i in range(0, nBins):
        histo[0][i] = histo[0][0] + i*dx
        histo[1][i] = 0.0
    return histo

def fillHisto(histo, x):
    L = histo[0].size
    dx = (histo[0][L-1]-histo[0][0])/(L-1.0)
    xMin = histo[0][0]-dx/2.0
    xMax = histo[0][L-1]+dx/2.0
    bin = binNb(xMin,xMax,L,x)
    if (0<=bin & bin<L):
        histo[1][bin]=histo[1][bin]+1

def normHisto(histo):
    L = histo[0].size
    sum = 0.0
    for i in range(0,L):
        sum = sum + histo[1][i]
    for i in range(0,L):
        histo[1][i] = histo[1][i]/sum

kahanSum = 0.0
kahanC = 0.0
xMin = truth-width
xMax = truth+width
histo = createHisto(xMin,xMax,nHistoBins)

for i in range(1, stat+1):
    x = x0 + eps*( 2.0*random.random() - 1.0)
    y = rn(13,x)

    kahanY = y - kahanC
    kahanT = kahanSum + kahanY
    kahanC = (kahanT-kahanSum) - kahanY
    kahanSum = kahanT

    fillHisto(histo,y)

single = rn(n, x0)
d_st = abs(single-truth)
avg = kahanSum/stat
d_av = abs(avg - truth)
d_rat = d_st / d_av

print("Truth = " + str(truth))
print("AV = " + str(avg) + ", delta = " + str(d_av))
print("Single = " + str(single) + ", delta = " + str(d_st))
print("Sing/AV delta  = " + str(d_rat))

normHisto(histo)
print()
for i in range (0, nHistoBins):
    xVal = str(histo[0][i])
    yVal = str(histo[1][i])
    print( "(" + xVal + "," + yVal + ") " , end="")
